/*
 * ODE-SOLVER numerical solution of ODE.
 *
 * Method Runge-Kutta of the fouth order. 
 * (http://en.wikipedia.org/wiki/Runge-Kutta_methods)
 *
 * Metod Runge-Kutta-Fehlberg.
 * (www.cnam.fr/math/IMG/pdf/RungeKuttaFehlbergProof.pdf)
 *
 * Copyright (c) 2010 Andrey Anufriev and Grigoriy Sitkarev
 * 
 * This software is distributed under the terms of GNU
 * Public License version 2 or later versions.
 *
 */
#ifndef ODE_SOLVER_H_
#define ODE_SOLVER_H_

enum {
	ODE_SOLVER_OK = 0,
	ODE_SOLVER_ERR = -1,
};

enum {
	ODE_SOLVER_H_ACCEPT,
	ODE_SOLVER_H_INCREASE,
	ODE_SOLVER_H_DECREASE,
};

enum {
	FALSE = 0,
	TRUE = 1,
};

struct ode_solver;

struct ode_solver_system;

typedef void (*ode_solver_deinit_func_t)(struct ode_solver *self);

typedef int (*ode_solver_step_func_t)(struct ode_solver *self, struct ode_solver_system *sys,
				      float x, float h, float y[], const float dy_in[],
				      float dy_out[], float y_err[]);

typedef void (*ode_solver_model_func_t)(float x, const float y[], float dy[], void *opaque);

typedef int (*ode_solver_modify_func_t)(struct ode_solver_system *sys, const float epsilon, float *h, const float y_err[]);

struct ode_solver {
	/* Solver method short name. */
	const char			*name;

	/* Solver order. */
	unsigned int			order;

	/* TRUE if method evaluates relative error, FALSE otherwise. */
	unsigned int			evaluates_yerr;

	/*
	 * Virtual functions.
	 */
	ode_solver_step_func_t		solver_step;
	ode_solver_deinit_func_t	solver_deinit;
};

struct ode_solver_system {
	/* System model evaluating derivatives using intial conditions. */
	ode_solver_model_func_t		model;

	/* Dimension of the 1-st order ODE system used by model. */
	unsigned int			dimension;

	/* Pointer to arbitrary data provided to model function. */
	void				*opaque;
};

/* Create RK4 ODE solver. */
struct ode_solver *ode_solver_rk4_new();

/* Create RKF45 ODE solver. */
struct ode_solver *ode_solver_rkf45_new();

/* Destroy ODE solver instance. */
void ode_solver_free(struct ode_solver *solver);

/* Perform integration step from x to x+h using solver `solver' and system `sys'. */
int ode_solver_step(struct ode_solver *solver, struct ode_solver_system *sys, float x, float h, float y[], const float dy_in[], float dy_out[], float y_err[]);

/* 
 * ode_solver_modify_h - step correction method.
 *
 * sys:      pointer to ode solver system instance
 * epsilon:  tolerance
 * h:        pointer to step
 * y_err:    relative error
 *
 * Returns: ODE_SOLVER_H_INCREASE if step increases
 *          ODE_SOLVER_H_DECREASE if step decreases
 *          error if invalid parameters
 *          ODE_SOLVER_OK else
 */

int ode_solver_modify_h(struct ode_solver_system *sys, const float epsilon, float *h, const float y_err[]);

#endif	/* ODE_SOLVER_H_ */

